import fnmatch
import re

from validator import decorator
from validator.contextgenerator import ContextGenerator
from validator.constants import PACKAGE_LANGPACK

BAD_LINK = '(href|src)=["\'](?!chrome:\/\/)(([a-z]*:)?\/\/|data:)'


@decorator.register_test(tier=2, expected_type=PACKAGE_LANGPACK)
def test_langpack_manifest(err, xpi_package=None):
    """Tests the chrome.manifest files in the package for
    compliance with the standard language pack triples."""

    # Don't even both with the test(s) if there's no chrome.manifest.
    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    for triple in chrome.triples:
        subject = triple['subject']
        # Test to make sure that the triple's subject is valid
        if subject not in ('locale', 'override', 'manifest'):
            err.warning(('testcases_langpack',
                         'test_langpack_manifest',
                         'invalid_subject'),
                        'Invalid chrome.manifest subject',
                        ['chrome.manifest files in language packs are only '
                         'allowed to contain items that are prefixed with '
                         "'locale', 'manifest', or 'override'. Other values "
                         'are not allowed.',
                         'Invalid subject: %s' % subject],
                        filename=triple['filename'],
                        line=triple['line'],
                        context=triple['context'])

        if subject == 'override':
            object_ = triple['object']
            predicate = triple['predicate']

            pattern = 'chrome://*/locale/*'

            if not fnmatch.fnmatch(object_, pattern) or \
               not fnmatch.fnmatch(predicate, pattern):
                err.warning(('testcases_langpack',
                             'test_langpack_manifest',
                             'invalid_override'),
                            'Invalid chrome.manifest object/predicate.',
                            "'override' entry does not match '%s'" % pattern,
                            filename=triple['filename'],
                            line=triple['line'],
                            context=triple['context'])


# This function is called by content.py
def test_unsafe_html(err, filename, data):
    'Tests for unsafe HTML tags in language pack files.'

    context = ContextGenerator(data)

    unsafe_pttrn = re.compile('<(script|embed|object)', re.I)

    match = unsafe_pttrn.search(data)
    if match:
        line = context.get_line(match.start())
        err.warning(('testcases_langpack',
                     'test_unsafe_html',
                     'unsafe_content_html'),
                    'Unsafe HTML found in language pack files.',
                    'Language packs are not allowed to contain scripts, '
                    'embeds, or other executable code in the language '
                    'definition files.',
                    filename,
                    line=line,
                    context=context)

    remote_pttrn = re.compile(BAD_LINK, re.I)

    match = remote_pttrn.search(data)
    if match:
        line = context.get_line(match.start())
        err.warning(('testcases_langpack',
                     'test_unsafe_html',
                     'unsafe_content_link'),
                    'Unsafe remote resource found in language pack.',
                    'Language packs are not allowed to contain references to '
                    'remote resources.',
                    filename,
                    line=line,
                    context=context)





from itertools import imap, repeat

from mock import patch

from validator.testcases import content

from tests.helper import MockXPI
from tests.js_helper import TestCase


class TestFunctionTraversal(TestCase):
    """
    Consider the following tree:

    - body
      |-function a()
      | |- foo
      | |- bar
      |-zip
      |-function b()
      | |-abc
      | |-def
      |-zap

    In the old traversal technique, it would be evaluated in the order:

        body a() foo bar zip b() abc def zap

    If the tree is considered as a graph, this would be prefix notation
    traversal.

    This is not optimal, however, as JS commonly uses callbacks which are set
    up before delegation code. The new traversal technique should access nodes
    in the following order:

        body zip zap a() foo bar b() abc def

    If the tree is considered a graph, this would be a custom prefix notation
    traversal where all non-function nodes are traversed before all function
    nodes.
    """

    def test_function_declaration_order(self):
        """Test that function declarations happen in the right time."""

        self.run_script("""
        foo = "first";
        function test() {
            foo = "second";
        }
        bar = foo;
        """)
        self.assert_var_eq('bar', 'first')
        self.assert_var_eq('foo', 'second')

    def test_function_expression_order(self):
        """Test that function expressions happen in the right time."""

        self.run_script("""
        foo = "first"
        var x = function() {
            foo = "second";
        }
        bar = foo;
        """)
        self.assert_var_eq('bar', 'first')
        self.assert_var_eq('foo', 'second')

    def test_nested_functions(self):
        """Test that nested functions are considered in the right order."""

        self.run_script("""
        foo = "first"
        function test() {
            function a() {foo = "second"}
            foo = "third";
        }
        """)
        self.assert_var_eq('foo', 'second')


class TestTooMuchJS(TestCase):
    """
    If you have an add-on and you trigger this code, then you're probably doing
    it wrong. This tests that any add-on with more than 1000 JS files in any
    given XPI or JAR do not perform exhaustive validation.
    """

    @patch('validator.testcases.content.run_regex_tests', lambda *a, **kw: None)
    @patch('validator.testcases.scripting.test_js_file', lambda *a, **kw: None)
    def run_mocked_scripts(self, count):
        self.setup_err()

        scripts = dict(zip(imap(lambda i: 's%d.js' % i, xrange(count)),
                           repeat('tests/resources/content/regex_error.js')))

        x = MockXPI(scripts)

        self.err.save_resource(
            'scripts',
            [{'scripts': scripts.keys(),
              'package': x,
              'state': []}])

        content.test_packed_scripts(self.err, x)

    def test_few_js(self):
        self.run_mocked_scripts(900)
        self.assert_silent()

    def test_many_js(self):
        self.run_mocked_scripts(1001)
        self.assert_failed(with_warnings=True)



from helper import CompatTestCase
from validator.compat import FX56_DEFINITION


class TestFX56Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 56 are properly executed."""

    VERSION = FX56_DEFINITION

    def test_showModalDialog_removed(self):
        """https://github.com/mozilla/amo-validator/issues/557"""
        expected = 'The showModalDialog function has been removed.'

        codes = [
          (
            'showModalDialog('
              '"http://developer.mozilla.org/samples/domref/showModalDialogBox.html",'
              '[1, 4], "dialogwidth: 450; dialogheight: 300; resizable: yes");'
          ),
          (
            'window.showModalDialog('
              '\'javascript:document.writeln("<script>alert(foo)" + "<" + "/script>")\');'
          )
        ]

        for code in codes:
          self.run_script_for_compat(code)

          assert not self.compat_err.notices
          assert not self.compat_err.errors

          self.assert_compat_error('warning', expected)



from zipfile import ZipFile
from StringIO import StringIO


def to_utf8(s):
    if isinstance(s, unicode):
        s = s.encode('utf-8')
    return s


class XPIManager(object):
    """
    An XPI reader and management class. Allows fun things like reading,
    listing, and extracting files from an XPI without you needing to
    worry about things like zip files or IO.
    """

    def __init__(self, package, mode='r', name=None, subpackage=False):
        'Create a new managed XPI package'

        self.zf = ZipFile(package, mode=mode)

        # Store away the filename for future use.
        self.filename = name or package
        self.extension = self.filename.split('.')[-1]
        self.subpackage = subpackage

        self.contents_cache = None

    def __iter__(self):
        return (name for name in self.zf.namelist())

    def __contains__(self, item):
        return item in self.zf.namelist()

    def info(self, name):
        """Get info on a single file."""
        return self.package_contents()[name]

    def package_contents(self):
        'Returns a dictionary of file information'

        if self.contents_cache:
            return self.contents_cache

        # Get a list of ZipInfo objects.
        files = self.zf.infolist()
        out_files = {}

        # Iterate through each file in the XPI.
        for file_ in files:

            file_doc = {'name': file_.filename,
                        'size': file_.file_size,
                        'name_lower': file_.filename.lower()}

            file_doc['extension'] = file_doc['name_lower'].split('.')[-1]

            out_files[file_.filename] = file_doc

        self.contents_cache = out_files
        return out_files

    def read(self, filename):
        'Reads a file from the archive and returns a string.'

        data = self.zf.read(filename)
        return data

    def write(self, name, data):
        """Write a blob of data to the XPI manager."""
        if isinstance(data, StringIO):
            self.zf.writestr(name, data.getvalue())
        else:
            self.zf.writestr(name, to_utf8(data))

    def write_file(self, name, path=None):
        """Write the contents of a file from the disk to the XPI."""

        if path is None:
            path = name

        self.zf.write(path, name)



import types
from StringIO import StringIO
from validator.contextgenerator import ContextGenerator


class PropertiesParser(object):
    """
    Parses and serializes .properties files. Even though you can pretty
    much do this in your sleep, it's still useful for L10n tests.
    """

    def __init__(self, dtd):
        """
        Properties parsers can initialized based on a file path
        (provided as a string to the path), or directly (in memory as a
        StringIO object).
        """

        self.entities = {}
        self.items = []

        if isinstance(dtd, types.StringTypes):
            data = open(dtd).read()
        elif isinstance(dtd, StringIO):
            data = dtd.getvalue()
        elif isinstance(dtd, file):
            data = dtd.read()

        # Create a context!
        self.context = ContextGenerator(data)

        split_data = data.split('\n')
        line_buffer = None
        line_number = 0
        for line in split_data:

            # Increment the line number
            line_number += 1

            # Clean things up
            clean_line = line.strip()
            if not clean_line:
                continue
            if clean_line.startswith('#'):
                continue

            # It's a line that wraps
            if clean_line.count('=') == 0:
                if line_buffer:
                    line_buffer[-1] += clean_line
                else:
                    continue
            else:

                if line_buffer:
                    # This line terminates a wrapped line
                    self.entities[line_buffer[0].strip()] = \
                        line_buffer[1].strip()
                    self.items.append((line_buffer[0].strip(),
                                       line_buffer[1].strip(),
                                       line_number))

                line_buffer = clean_line.split('=', 1)

        # Handle any left-over wrapped line data
        if line_buffer:
            self.entities[line_buffer[0].strip()] = \
                line_buffer[1].strip()
            self.items.append((line_buffer[0].strip(),
                               line_buffer[1].strip(),
                               line_number))

    def __len__(self):
        return len(self.entities)




import logging

import validator.constants
from validator.constants import APPLICATIONS


log = logging.getLogger('amo.validator')


TEST_TIERS = {}
CLEANUP_FUNCTIONS = []


def register_test(tier=1, expected_type=None, simple=False, versions=None):
    'Registers tests for the validation flow.'

    def wrap(function):
        'Wrapper function to decorate registered tests.'

        # Make sure the tier exists before we add to it
        if tier not in TEST_TIERS:
            TEST_TIERS[tier] = []

        # Add a test object to the test's tier
        TEST_TIERS[tier].append({'test': function,
                                 'type': expected_type,
                                 'simple': simple,
                                 'versions': versions})

        # Return the function to be run
        return function

    # Return the wrapping function (for use as a decorator)
    return wrap


def register_cleanup(cleanup):
    """Register a cleanup function to be called at the end of every validation
    task. Takes either a callable (including a class with a __call_ method),
    or a class with a `cleanup` class method."""

    if not callable(cleanup):
        # Allow decorating a class with a `cleanup` classm ethod.
        cleanup = cleanup.cleanup

    CLEANUP_FUNCTIONS.append(cleanup.cleanup)
    return cleanup


def cleanup():
    """Call every cleanup function which has been registered via
    @register_cleanup."""

    for fn in CLEANUP_FUNCTIONS:
        try:
            fn()
        except Exception, e:
            log.exception('Error during cleanup: %s' % e)


def get_tiers():
    'Returns a list of tier values.'
    return TEST_TIERS.keys()


def get_tests(tier, type_=None):
    'Returns a generator of test functions.'

    # List of acceptable types
    types = (None, 0, type_)

    # Current Tier
    ctier = TEST_TIERS[tier]

    # List comprehension to sort and filter and the like.
    return (test for test in ctier if test['type'] in types)


def version_range(guid, version, before=None, app_versions=None):
    """Returns all values after (and including) `version` for the app `guid`"""

    if app_versions is None:
        app_versions = validator.constants.APPROVED_APPLICATIONS
    app_key = None

    # Support for shorthand instead of full GUIDs.
    for app_guid, app_name in APPLICATIONS.items():
        if app_name == guid:
            guid = app_guid
            break

    for key in app_versions.keys():
        if app_versions[key]['guid'] == guid:
            app_key = key
            break

    if not app_key or version not in app_versions[app_key]['versions']:
        raise Exception('Bad GUID or version provided for version range: %s'
                        % version)

    all_versions = app_versions[app_key]['versions']
    version_pos = all_versions.index(version)
    before_pos = None
    if before is not None and before in all_versions:
        before_pos = all_versions.index(before)

    return all_versions[version_pos:before_pos]



from copy import deepcopy
from functools import partial
import sys
import types

# Global import of predefinedentities will cause an import loop
import instanceactions
from validator.constants import (BUGZILLA_BUG, DESCRIPTION_TYPES, FENNEC_GUID,
                                 FIREFOX_GUID, MAX_STR_SIZE, MDN_DOC)
from validator.decorator import version_range
from jstypes import JSArray, JSContext, JSLiteral, JSObject, JSWrapper


NUMERIC_TYPES = (int, long, float, complex)

# None of these operations (or their augmented assignment counterparts) should
# be performed on non-numeric data. Any time we get non-numeric data for these
# guys, we just return window.NaN.
NUMERIC_OPERATORS = ('-', '*', '/', '%', '<<', '>>', '>>>', '|', '^', '&')
NUMERIC_OPERATORS += tuple('%s=' % op for op in NUMERIC_OPERATORS)


def get_NaN(traverser):
    # If we've cached the traverser's NaN instance, just use that.
    ncache = getattr(traverser, 'NAN_CACHE', None)
    if ncache is not None:
        return ncache

    # Otherwise, we need to import GLOBAL_ENTITIES and build a raw copy.
    from predefinedentities import GLOBAL_ENTITIES
    ncache = traverser._build_global('NaN', GLOBAL_ENTITIES[u'NaN'])
    # Cache it so we don't need to do this again.
    traverser.NAN_CACHE = ncache
    return ncache


def _get_member_exp_property(traverser, node):
    """Return the string value of a member expression's property."""

    if node['property']['type'] == 'Identifier' and not node.get('computed'):
        return unicode(node['property']['name'])
    else:
        eval_exp = traverser._traverse_node(node['property'])
        return _get_as_str(eval_exp.get_literal_value())


def _expand_globals(traverser, node):
    """Expands a global object that has a lambda value."""

    if node.is_global and callable(node.value.get('value')):
        result = node.value['value'](traverser)
        if isinstance(result, dict):
            output = traverser._build_global('--', result)
        elif isinstance(result, JSWrapper):
            output = result
        else:
            output = JSWrapper(result, traverser)

        # Set the node context.
        if 'context' in node.value:
            traverser._debug('CONTEXT>>%s' % node.value['context'])
            output.context = node.value['context']
        else:
            traverser._debug('CONTEXT>>INHERITED')
            output.context = node.context

        return output

    return node


def trace_member(traverser, node, instantiate=False):
    'Traces a MemberExpression and returns the appropriate object'

    traverser._debug('TESTING>>%s' % node['type'])
    if node['type'] == 'MemberExpression':
        # x.y or x[y]
        # x = base
        base = trace_member(traverser, node['object'], instantiate)
        base = _expand_globals(traverser, base)

        identifier = _get_member_exp_property(traverser, node)

        # Handle the various global entity properties.
        if base.is_global:
            # If we've got an XPCOM wildcard, return a copy of the entity.
            if 'xpcom_wildcard' in base.value:
                traverser._debug('MEMBER_EXP>>XPCOM_WILDCARD')

                from predefinedentities import CONTRACT_ENTITIES
                if identifier in CONTRACT_ENTITIES:
                    kw = dict(err_id=('js', 'actions', 'dangerous_contract'),
                              warning='Dangerous XPCOM contract ID')
                    kw.update(CONTRACT_ENTITIES[identifier])

                    traverser.warning(**kw)

                base.value = base.value.copy()
                del base.value['xpcom_wildcard']
                return base

        test_identifier(traverser, identifier)

        traverser._debug('MEMBER_EXP>>PROPERTY: %s' % identifier)
        output = base.get(
            traverser=traverser, instantiate=instantiate, name=identifier)
        output.context = base.context

        if base.is_global:
            # In the cases of XPCOM objects, methods generally
            # remain bound to their parent objects, even when called
            # indirectly.
            output.parent = base
        return output

    elif node['type'] == 'Identifier':
        traverser._debug('MEMBER_EXP>>ROOT:IDENTIFIER')
        test_identifier(traverser, node['name'])

        # If we're supposed to instantiate the object and it doesn't already
        # exist, instantitate the object.
        if instantiate and not traverser._is_defined(node['name']):
            output = JSWrapper(JSObject(), traverser=traverser)
            traverser.contexts[0].set(node['name'], output)
        else:
            output = traverser._seek_variable(node['name'])

        return _expand_globals(traverser, output)
    else:
        traverser._debug('MEMBER_EXP>>ROOT:EXPRESSION')
        # It's an expression, so just try your damndest.
        return traverser._traverse_node(node)


def test_identifier(traverser, name):
    'Tests whether an identifier is banned'

    import predefinedentities
    if name in predefinedentities.BANNED_IDENTIFIERS:
        traverser.err.warning(
            err_id=('js', 'actions', 'banned_identifier'),
            warning='Banned or deprecated JavaScript Identifier',
            description=predefinedentities.BANNED_IDENTIFIERS[name],
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)


def _function(traverser, node):
    'Prevents code duplication'

    def wrap(traverser, node):
        me = JSObject()

        traverser.function_collection.append([])

        # Replace the current context with a prototypeable JS object.
        traverser._pop_context()
        me.type_ = 'default'  # Treat the function as a normal object.
        traverser._push_context(me)
        traverser._debug('THIS_PUSH')
        traverser.this_stack.append(me)  # Allow references to "this"

        # Declare parameters in the local scope
        params = []
        for param in node['params']:
            if param['type'] == 'Identifier':
                params.append(param['name'])
            elif param['type'] == 'ArrayPattern':
                for element in param['elements']:
                    # Array destructuring in function prototypes? LOL!
                    if element is None or element['type'] != 'Identifier':
                        continue
                    params.append(element['name'])

        local_context = traverser._peek_context(1)
        for param in params:
            var = JSWrapper(lazy=True, traverser=traverser)

            # We can assume that the params are static because we don't care
            # about what calls the function. We want to know whether the
            # function solely returns static values. If so, it is a static
            # function.
            local_context.set(param, var)

        traverser._traverse_node(node['body'])

        # Since we need to manually manage the "this" stack, pop off that
        # context.
        traverser._debug('THIS_POP')
        traverser.this_stack.pop()

        # Call all of the function collection's members to traverse all of the
        # child functions.
        func_coll = traverser.function_collection.pop()
        for func in func_coll:
            func()

    # Put the function off for traversal at the end of the current block scope.
    traverser.function_collection[-1].append(partial(wrap, traverser, node))

    return JSWrapper(traverser=traverser, callable=True, dirty=True)


def _define_function(traverser, node):
    me = _function(traverser, node)
    traverser._peek_context(2).set(node['id']['name'], me)
    return me


def _func_expr(traverser, node):
    'Represents a lambda function'

    return _function(traverser, node)


def _define_with(traverser, node):
    'Handles `with` statements'

    object_ = traverser._traverse_node(node['object'])
    if isinstance(object_, JSWrapper) and isinstance(object_.value, JSObject):
        traverser.contexts[-1] = object_.value
        traverser.contexts.append(JSContext('block'))
    return


def _define_var(traverser, node):
    'Creates a local context variable'

    traverser._debug('VARIABLE_DECLARATION')
    traverser.debug_level += 1

    declarations = (node['declarations'] if 'declarations' in node
                    else node['head'])

    kind = node.get('kind', 'let')
    for declaration in declarations:

        # It could be deconstruction of variables :(
        if declaration['id']['type'] == 'ArrayPattern':

            vars = []
            for element in declaration['id']['elements']:
                # NOTE : Multi-level array destructuring sucks. Maybe implement
                # it someday if you're bored, but it's so rarely used and it's
                # so utterly complex, there's probably no need to ever code it
                # up.
                if element is None or element['type'] != 'Identifier':
                    vars.append(None)
                    continue
                vars.append(element['name'])

            # The variables are not initialized
            if declaration['init'] is None:
                # Simple instantiation; no initialization
                for var in vars:
                    if not var:
                        continue
                    traverser._declare_variable(var, None)

            # The variables are declared inline
            elif declaration['init']['type'] == 'ArrayPattern':
                # TODO : Test to make sure len(values) == len(vars)
                for value in declaration['init']['elements']:
                    if vars[0]:
                        traverser._declare_variable(
                            vars[0], JSWrapper(traverser._traverse_node(value),
                                               traverser=traverser))
                    vars = vars[1:]  # Pop off the first value

            # It's being assigned by a JSArray (presumably)
            elif declaration['init']['type'] == 'ArrayExpression':

                assigner = traverser._traverse_node(declaration['init'])
                for value in assigner.value.elements:
                    if vars[0]:
                        traverser._declare_variable(vars[0], value)
                    vars = vars[1:]

        elif declaration['id']['type'] == 'ObjectPattern':

            init = traverser._traverse_node(declaration['init'])

            def _proc_objpattern(init_obj, properties):
                for prop in properties:
                    # Get the name of the init obj's member
                    if prop['key']['type'] == 'Literal':
                        prop_name = prop['key']['value']
                    elif prop['key']['type'] == 'Identifier':
                        prop_name = prop['key']['name']
                    else:
                        continue

                    if prop['value']['type'] == 'Identifier':
                        traverser._declare_variable(
                            prop['value']['name'],
                            init_obj.get(traverser, prop_name))
                    elif prop['value']['type'] == 'ObjectPattern':
                        _proc_objpattern(init_obj.get(traverser, prop_name),
                                         prop['value']['properties'])

            if init is not None:
                _proc_objpattern(init_obj=init,
                                 properties=declaration['id']['properties'])

        else:
            var_name = declaration['id']['name']
            traverser._debug('NAME>>%s' % var_name)

            var_value = traverser._traverse_node(declaration['init'])
            traverser._debug('VALUE>>%s' % (var_value.output()
                                            if var_value is not None
                                            else 'None'))

            if not isinstance(var_value, JSWrapper):
                var = JSWrapper(value=var_value,
                                const=kind == 'const',
                                traverser=traverser)
            else:
                var = var_value
                var.const = kind == 'const'

            traverser._declare_variable(var_name, var, type_=kind)

    if 'body' in node:
        traverser._traverse_node(node['body'])

    traverser.debug_level -= 1

    # The "Declarations" branch contains custom elements.
    return True


def _define_obj(traverser, node):
    'Creates a local context object'

    var = JSObject()
    for prop in node['properties']:
        if prop['type'] == 'PrototypeMutation':
            var_name = 'prototype'
        else:
            key = prop['key']
            if key['type'] == 'Literal':
                var_name = key['value']
            elif isinstance(key['name'], basestring):
                var_name = key['name']
            else:
                if 'property' in key['name']:
                    name = key['name']
                else:
                    name = {'property': key['name']}
                var_name = _get_member_exp_property(traverser, name)

        var_value = traverser._traverse_node(prop['value'])
        var.set(var_name, var_value, traverser)

        # TODO: Observe "kind"

    if not isinstance(var, JSWrapper):
        return JSWrapper(var, lazy=True, traverser=traverser)
    var.lazy = True
    return var


def _define_array(traverser, node):
    """Instantiate an array object from the parse tree."""
    arr = JSArray()
    arr.elements = map(traverser._traverse_node, node['elements'])
    return arr


def _define_template_strings(traverser, node):
    """Instantiate an array of raw and cooked template strings."""
    cooked = JSArray()
    cooked.elements = map(traverser._traverse_node, node['cooked'])
    raw = JSArray()
    raw.elements = map(traverser._traverse_node, node['raw'])

    cooked.set('raw', raw, traverser)
    return cooked


def _define_template(traverser, node):
    """Instantiate a template literal."""
    elements = map(traverser._traverse_node, node['elements'])

    return reduce(partial(_binary_op, '+', traverser=traverser), elements)


def _define_literal(traverser, node):
    """
    Convert a literal node in the parse tree to its corresponding
    interpreted value.
    """
    value = node['value']
    if isinstance(value, dict):
        return JSWrapper(JSObject(), traverser=traverser, dirty=True)

    wrapper = JSWrapper(value if value is not None else JSLiteral(None),
                        traverser=traverser)
    test_literal(traverser, wrapper)
    return wrapper


def test_literal(traverser, wrapper):
    """
    Test the value of a literal, in particular only a string literal at the
    moment, against possibly dangerous patterns.
    """
    value = wrapper.get_literal_value()
    if isinstance(value, basestring):
        # Local import to prevent import loop.
        from validator.testcases.regex import validate_string
        validate_string(value, traverser, wrapper=wrapper)


def _call_expression(traverser, node):
    args = node['arguments']
    for arg in args:
        traverser._traverse_node(arg, source='arguments')

    member = traverser._traverse_node(node['callee'])

    if (traverser.filename.startswith('defaults/preferences/') and
        ('name' not in node['callee'] or
         node['callee']['name'] not in (u'pref', u'user_pref'))):

        traverser.err.warning(
            err_id=('testcases_javascript_actions',
                    '_call_expression',
                    'complex_prefs_defaults_code'),
            warning='Complex code should not appear in preference defaults '
                    'files',
            description="Calls to functions other than 'pref' and 'user_pref' "
                        'should not appear in defaults/preferences/ files.',
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)

    if member.is_global and callable(member.value.get('dangerous', None)):
        result = member.value['dangerous'](a=args, t=traverser._traverse_node,
                                           e=traverser.err)
        name = member.value.get('name', '')

        if result and name:
            kwargs = {
                'err_id': ('testcases_javascript_actions', '_call_expression',
                           'called_dangerous_global'),
                'warning': '`%s` called in potentially dangerous manner' %
                           member.value['name'],
                'description':
                    'The global `%s` function was called using a set '
                    'of dangerous parameters. Calls of this nature '
                    'are deprecated.' % member.value['name']}

            if isinstance(result, DESCRIPTION_TYPES):
                kwargs['description'] = result
            elif isinstance(result, dict):
                kwargs.update(result)

            traverser.warning(**kwargs)

    elif (node['callee']['type'] == 'MemberExpression' and
          node['callee']['property']['type'] == 'Identifier'):

        # If we can identify the function being called on any member of any
        # instance, we can use that to either generate an output value or test
        # for additional conditions.
        identifier_name = node['callee']['property']['name']
        if identifier_name in instanceactions.INSTANCE_DEFINITIONS:
            result = instanceactions.INSTANCE_DEFINITIONS[identifier_name](
                        args, traverser, node, wrapper=member)
            return result

    if member.is_global and 'return' in member.value:
        if 'object' in node['callee']:
            member.parent = trace_member(traverser, node['callee']['object'])

        return member.value['return'](wrapper=member, arguments=args,
                                      traverser=traverser)
    return JSWrapper(JSObject(), dirty=True, traverser=traverser)


def _call_settimeout(a, t, e):
    """
    Handler for setTimeout and setInterval. Should determine whether a[0]
    is a lambda function or a string. Strings are banned, lambda functions are
    ok. Since we can't do reliable type testing on other variables, we flag
    those, too.
    """

    if not a:
        return

    if a[0]['type'] in ('FunctionExpression', 'ArrowFunctionExpression'):
        return

    if t(a[0]).callable:
        return

    return {'err_id': ('javascript', 'dangerous_global', 'eval'),
            'description':
                'In order to prevent vulnerabilities, the `setTimeout` '
                'and `setInterval` functions should be called only with '
                'function expressions as their first argument.',
            'signing_help': (
                'Please do not ever call `setTimeout` or `setInterval` with '
                'string arguments. If you are passing a function which is '
                'not being correctly detected as such, please consider '
                'passing a closure or arrow function, which in turn calls '
                'the original function.'),
            'signing_severity': 'high'}


def _call_require(a, t, e):
    """
    Tests for unsafe uses of `require()` in SDK add-ons.
    """

    args, traverse, err = a, t, e

    if not err.metadata.get('is_jetpack') and len(args):
        return

    module = traverse(args[0]).get_literal_value()
    if not isinstance(module, basestring):
        return

    if module.startswith('sdk/'):
        module = module[len('sdk/'):]

    LOW_LEVEL = {
        # Added from bugs 689340, 731109
        'chrome', 'window-utils', 'observer-service',
        # Added from bug 845492
        'window/utils', 'sdk/window/utils', 'sdk/deprecated/window-utils',
        'tab/utils', 'sdk/tab/utils',
        'system/events', 'sdk/system/events',
    }

    if module in LOW_LEVEL:
        err.metadata['requires_chrome'] = True
        return {'warning': 'Usage of low-level or non-SDK interface',
                'description': 'Your add-on uses an interface which bypasses '
                               'the high-level protections of the add-on SDK. '
                               'This interface should be avoided, and its use '
                               'may significantly complicate your review '
                               'process.'}

    if module == 'widget':
        return {'warning': 'Use of deprecated SDK module',
                'description':
                    "The 'widget' module has been deprecated due to a number "
                    'of performance and usability issues, and has been '
                    'removed from the SDK as of Firefox 40. Please use the '
                    "'sdk/ui/button/action' or 'sdk/ui/button/toggle' module "
                    'instead. See '
                    'https://developer.mozilla.org/Add-ons/SDK/High-Level_APIs'
                    '/ui for more information.'}


def _call_create_pref(a, t, e):
    """
    Handler for pref() and user_pref() calls in defaults/preferences/*.js files
    to ensure that they don't touch preferences outside of the "extensions."
    branch.
    """

    # We really need to clean up the arguments passed to these functions.
    traverser = t.im_self

    if not traverser.filename.startswith('defaults/preferences/') or not a:
        return

    instanceactions.set_preference(JSWrapper(JSLiteral(None),
                                             traverser=traverser),
                                   a, traverser)

    value = _get_as_str(t(a[0]))
    return test_preference(value)


def test_preference(value):
    for branch in 'extensions.', 'services.sync.prefs.sync.extensions.':
        if value.startswith(branch) and value.rindex('.') > len(branch):
            return

    return ('Extensions should not alter preferences outside of the '
            "'extensions.' preference branch. Please make sure that "
            "all of your extension's preferences are prefixed with "
            "'extensions.add-on-name.', where 'add-on-name' is a "
            'distinct string unique to and indicative of your add-on.')


def _expression(traverser, node):
    """
    This is a helper method that allows node definitions to point at
    `_traverse_node` without needing a reference to a traverser.
    """
    return traverser._traverse_node(node['expression'])


def _get_this(traverser, node):
    'Returns the `this` object'
    if not traverser.this_stack:
        from predefinedentities import GLOBAL_ENTITIES
        return traverser._build_global('window', GLOBAL_ENTITIES[u'window'])
    return traverser.this_stack[-1]


def _new(traverser, node):
    'Returns a new copy of a node.'

    # We don't actually process the arguments as part of the flow because of
    # the Angry T-Rex effect. For now, we just traverse them to ensure they
    # don't contain anything dangerous.
    args = node['arguments']
    if isinstance(args, list):
        for arg in args:
            traverser._traverse_node(arg, source='arguments')
    else:
        traverser._traverse_node(args)

    elem = traverser._traverse_node(node['callee'])
    if not isinstance(elem, JSWrapper):
        elem = JSWrapper(elem, traverser=traverser)
    if elem.is_global:
        traverser._debug('Making overwritable')
        elem.value = deepcopy(elem.value)
        elem.value['overwritable'] = True
    return elem


def _ident(traverser, node):
    'Initiates an object lookup on the traverser based on an identifier token'

    name = node['name']

    # Ban bits like "newThread"
    test_identifier(traverser, name)

    if traverser._is_defined(name):
        return traverser._seek_variable(name)

    return JSWrapper(JSObject(), traverser=traverser, dirty=True)


def _expr_assignment(traverser, node):
    """Evaluate an AssignmentExpression node."""

    traverser._debug('ASSIGNMENT_EXPRESSION')
    traverser.debug_level += 1

    traverser._debug('ASSIGNMENT>>PARSING RIGHT')
    right = traverser._traverse_node(node['right'])
    right = JSWrapper(right, traverser=traverser)

    # Treat direct assignment different than augmented assignment.
    if node['operator'] == '=':
        from predefinedentities import GLOBAL_ENTITIES, is_shared_scope

        global_overwrite = False
        readonly_value = is_shared_scope(traverser)

        node_left = node['left']
        traverser._debug('ASSIGNMENT:DIRECT(%s)' % node_left['type'])

        if node_left['type'] == 'Identifier':
            # Identifiers just need the ID name and a value to push.
            # Raise a global overwrite issue if the identifier is global.
            global_overwrite = traverser._is_global(node_left['name'])

            # Get the readonly attribute and store its value if is_global
            if global_overwrite:
                global_dict = GLOBAL_ENTITIES[node_left['name']]
                if 'readonly' in global_dict:
                    readonly_value = global_dict['readonly']

            traverser._declare_variable(node_left['name'], right, type_='glob')
        elif node_left['type'] == 'MemberExpression':
            member_object = trace_member(traverser, node_left['object'],
                                         instantiate=True)
            global_overwrite = (member_object.is_global and
                                not ('overwritable' in member_object.value and
                                     member_object.value['overwritable']))
            member_property = _get_member_exp_property(traverser, node_left)
            traverser._debug('ASSIGNMENT:MEMBER_PROPERTY(%s)'
                             % member_property)
            traverser._debug('ASSIGNMENT:GLOB_OV::%s' % global_overwrite)

            # Don't do the assignment if we're facing a global.
            if not member_object.is_global:
                if member_object.value is None:
                    member_object.value = JSObject()

                if not member_object.is_global:
                    member_object.value.set(member_property, right, traverser)
                else:
                    # It's probably better to do nothing.
                    pass

            elif 'value' in member_object.value:
                member_object_value = _expand_globals(traverser,
                                                      member_object).value
                if member_property in member_object_value['value']:

                    # If it's a global and the actual member exists, test
                    # whether it can be safely overwritten.
                    member = member_object_value['value'][member_property]
                    if 'readonly' in member:
                        global_overwrite = True
                        readonly_value = member['readonly']

        traverser._debug('ASSIGNMENT:DIRECT:GLOB_OVERWRITE %s' %
                         global_overwrite)
        traverser._debug('ASSIGNMENT:DIRECT:READONLY %r' %
                         readonly_value)

        if callable(readonly_value):
            readonly_value = readonly_value(traverser, right, node['right'])

        if readonly_value and global_overwrite:

            kwargs = dict(
                err_id=('testcases_javascript_actions',
                        '_expr_assignment',
                        'global_overwrite'),
                warning='Global variable overwrite',
                description='An attempt was made to overwrite a global '
                            'variable in some JavaScript code.')

            if isinstance(readonly_value, DESCRIPTION_TYPES):
                kwargs['description'] = readonly_value
            elif isinstance(readonly_value, dict):
                kwargs.update(readonly_value)

            traverser.warning(**kwargs)

        return right

    lit_right = right.get_literal_value()

    traverser._debug('ASSIGNMENT>>PARSING LEFT')
    left = traverser._traverse_node(node['left'])
    traverser._debug('ASSIGNMENT>>DONE PARSING LEFT')
    traverser.debug_level -= 1

    if isinstance(left, JSWrapper):
        if left.dirty:
            return left

        lit_left = left.get_literal_value()
        token = node['operator']

        # Don't perform an operation on None. Python freaks out
        if lit_left is None:
            lit_left = 0
        if lit_right is None:
            lit_right = 0

        # Give them default values so we have them in scope.
        gleft, gright = 0, 0

        # All of the assignment operators
        operators = {'=': lambda: right,
                     '+=': lambda: lit_left + lit_right,
                     '-=': lambda: gleft - gright,
                     '*=': lambda: gleft * gright,
                     '/=': lambda: 0 if gright == 0 else (gleft / gright),
                     '%=': lambda: 0 if gright == 0 else (gleft % gright),
                     '<<=': lambda: int(gleft) << int(gright),
                     '>>=': lambda: int(gleft) >> int(gright),
                     '>>>=': lambda: float(abs(int(gleft)) >> gright),
                     '|=': lambda: int(gleft) | int(gright),
                     '^=': lambda: int(gleft) ^ int(gright),
                     '&=': lambda: int(gleft) & int(gright)}

        # If we're modifying a non-numeric type with a numeric operator, return
        # NaN.
        if (not isinstance(lit_left, NUMERIC_TYPES) and
                token in NUMERIC_OPERATORS):
            left.set_value(get_NaN(traverser), traverser=traverser)
            return left

        # If either side of the assignment operator is a string, both sides
        # need to be casted to strings first.
        if (isinstance(lit_left, types.StringTypes) or
                isinstance(lit_right, types.StringTypes)):
            lit_left = _get_as_str(lit_left)
            lit_right = _get_as_str(lit_right)

        gleft, gright = _get_as_num(left), _get_as_num(right)

        traverser._debug('ASSIGNMENT>>OPERATION:%s' % token)
        if token not in operators:
            # We don't support that operator. (yet?)
            traverser._debug('ASSIGNMENT>>OPERATOR NOT FOUND', 1)
            return left
        elif token in ('<<=', '>>=', '>>>=') and gright < 0:
            # The user is doing weird bitshifting that will return 0 in JS but
            # not in Python.
            left.set_value(0, traverser=traverser)
            return left
        elif (token in ('<<=', '>>=', '>>>=', '|=', '^=', '&=') and
              (abs(gleft) == float('inf') or abs(gright) == float('inf'))):
            # Don't bother handling infinity for integer-converted operations.
            left.set_value(get_NaN(traverser), traverser=traverser)
            return left

        traverser._debug('ASSIGNMENT::L-value global? (%s)' %
                         ('Y' if left.is_global else 'N'), 1)
        try:
            new_value = operators[token]()
        except Exception:
            traverser.system_error(exc_info=sys.exc_info())
            new_value = None

        # Cap the length of analyzed strings.
        if (isinstance(new_value, types.StringTypes) and
                len(new_value) > MAX_STR_SIZE):
            new_value = new_value[:MAX_STR_SIZE]

        traverser._debug('ASSIGNMENT::New value >> %s' % new_value, 1)
        left.set_value(new_value, traverser=traverser)
        return left

    # Though it would otherwise be a syntax error, we say that 4=5 should
    # evaluate out to 5.
    return right


def _expr_binary(traverser, node):
    'Evaluates a BinaryExpression node.'

    traverser.debug_level += 1

    # Select the proper operator.
    operator = node['operator']
    traverser._debug('BIN_OPERATOR>>%s' % operator)

    # Traverse the left half of the binary expression.
    with traverser._debug('BIN_EXP>>l-value'):
        if (node['left']['type'] == 'BinaryExpression' and
                '__traversal' not in node['left']):
            # Process the left branch of the binary expression directly. This
            # keeps the recursion cap in line and speeds up processing of
            # large chains of binary expressions.
            left = _expr_binary(traverser, node['left'])
            node['left']['__traversal'] = left
        else:
            left = traverser._traverse_node(node['left'])

    # Traverse the right half of the binary expression.
    with traverser._debug('BIN_EXP>>r-value'):
        if (operator == 'instanceof' and
                node['right']['type'] == 'Identifier' and
                node['right']['name'] == 'Function'):
            # We make an exception for instanceof's r-value if it's a
            # dangerous global, specifically Function.
            return JSWrapper(True, traverser=traverser)
        else:
            right = traverser._traverse_node(node['right'])
            traverser._debug('Is dirty? %r' % right.dirty, 1)

    return _binary_op(operator, left, right, traverser)


def _binary_op(operator, left, right, traverser):
    """Perform a binary operation on two pre-traversed nodes."""

    # Dirty l or r values mean we can skip the expression. A dirty value
    # indicates that a lazy operation took place that introduced some
    # nondeterminacy.
    # FIXME(Kris): We should process these as if they're strings anyway.
    if left.dirty:
        return left
    elif right.dirty:
        return right

    # Binary expressions are only executed on literals.
    left = left.get_literal_value()
    right_wrap = right
    right = right.get_literal_value()

    # Coerce the literals to numbers for numeric operations.
    gleft = _get_as_num(left)
    gright = _get_as_num(right)

    operators = {
        '==': lambda: left == right or gleft == gright,
        '!=': lambda: left != right,
        '===': lambda: left == right,  # Be flexible.
        '!==': lambda: type(left) != type(right) or left != right,
        '>': lambda: left > right,
        '<': lambda: left < right,
        '<=': lambda: left <= right,
        '>=': lambda: left >= right,
        '<<': lambda: int(gleft) << int(gright),
        '>>': lambda: int(gleft) >> int(gright),
        '>>>': lambda: float(abs(int(gleft)) >> int(gright)),
        '+': lambda: left + right,
        '-': lambda: gleft - gright,
        '*': lambda: gleft * gright,
        '/': lambda: 0 if gright == 0 else (gleft / gright),
        '%': lambda: 0 if gright == 0 else (gleft % gright),
        'in': lambda: right_wrap.contains(left),
        # TODO : implement instanceof
        # FIXME(Kris): Treat instanceof the same as `QueryInterface`
    }

    output = None
    if (operator in ('>>', '<<', '>>>') and
            (left is None or right is None or gright < 0)):
        output = False
    elif operator in operators:
        # Concatenation can be silly, so always turn undefineds into empty
        # strings and if there are strings, make everything strings.
        if operator == '+':
            if left is None:
                left = ''
            if right is None:
                right = ''
            if isinstance(left, basestring) or isinstance(right, basestring):
                left = _get_as_str(left)
                right = _get_as_str(right)

        # Don't even bother handling infinity if it's a numeric computation.
        if (operator in ('<<', '>>', '>>>') and
                (abs(gleft) == float('inf') or abs(gright) == float('inf'))):
            return get_NaN(traverser)

        try:
            output = operators[operator]()
        except Exception:
            traverser.system_error(exc_info=sys.exc_info())
            output = None

        # Cap the length of analyzed strings.
        if (isinstance(output, types.StringTypes) and
                len(output) > MAX_STR_SIZE):
            output = output[:MAX_STR_SIZE]

        wrapper = JSWrapper(output, traverser=traverser)

        # Test the newly-created literal for dangerous values.
        # This may cause duplicate warnings for strings which
        # already match a dangerous value prior to concatenation.
        test_literal(traverser, wrapper)

        return wrapper

    return JSWrapper(output, traverser=traverser)


def _expr_unary(traverser, node):
    """Evaluate a UnaryExpression node."""

    expr = traverser._traverse_node(node['argument'])
    expr_lit = expr.get_literal_value()
    expr_num = _get_as_num(expr_lit)

    operators = {'-': lambda: -1 * expr_num,
                 '+': lambda: expr_num,
                 '!': lambda: not expr_lit,
                 '~': lambda: -1 * (expr_num + 1),
                 'void': lambda: None,
                 'typeof': lambda: _expr_unary_typeof(expr),
                 'delete': lambda: None}  # We never want to empty the context
    if node['operator'] in operators:
        output = operators[node['operator']]()
    else:
        output = None

    if not isinstance(output, JSWrapper):
        output = JSWrapper(output, traverser=traverser)
    return output


def _expr_unary_typeof(wrapper):
    """Evaluate the "typeof" value for a JSWrapper object."""
    if (wrapper.callable or
        (wrapper.is_global and 'return' in wrapper.value and
         'value' not in wrapper.value)):
        return 'function'

    value = wrapper.value
    if value is None:
        return 'undefined'
    elif isinstance(value, JSLiteral):
        value = value.value
        if isinstance(value, bool):
            return 'boolean'
        elif isinstance(value, (int, long, float)):
            return 'number'
        elif isinstance(value, types.StringTypes):
            return 'string'
    return 'object'


def _get_as_num(value):
    """Return the JS numeric equivalent for a value."""

    if isinstance(value, JSWrapper):
        value = value.get_literal_value()
    if value is None:
        return 0

    try:
        if isinstance(value, types.StringTypes):
            if value.startswith('0x'):
                return int(value, 16)
            else:
                return float(value)
        elif isinstance(value, (int, float, long)):
            return value
        else:
            return int(value)
    except (ValueError, TypeError):
        return 0


def _get_as_str(value):
    """Return the JS string equivalent for a literal value."""

    if isinstance(value, JSWrapper):
        value = value.get_literal_value()
    if value is None:
        return ''

    if isinstance(value, bool):
        return u'true' if value else u'false'
    elif isinstance(value, (int, float, long)):
        if value == float('inf'):
            return u'Infinity'
        elif value == float('-inf'):
            return u'-Infinity'

        # Try to see if we can shave off some trailing significant figures.
        try:
            if int(value) == value:
                return unicode(int(value))
        except ValueError:
            pass
    return unicode(value)



from helper import CompatTestCase
from validator.compat import FX54_DEFINITION


class TestFX54Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 54 are properly executed."""

    VERSION = FX54_DEFINITION

    def test_formhistory2_removed(self):
        """https://github.com/mozilla/amo-validator/issues/523"""
        expected = (
            'The nsIFormHistory2 interface has been removed. You can use '
            'FormHistory.jsm instead.')

        code = '''
            Components.classes["@mozilla.org/satchel/form-history;1"].getService(Components.interfaces.nsIFormHistory2)
        '''

        self.run_script_for_compat(code)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

    def test_moz_appearance(self):
        """https://github.com/mozilla/amo-validator/issues/524"""
        expected = '-moz-appearance can only be used in chrome stylesheets.'

        invalid = '-moz-border-radius: 0px; -moz-appearance:textfield;'

        self.run_regex_for_compat(invalid, is_js=False)
        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_warning('warning', expected)



from helper import CompatTestCase
from validator.compat import FX45_DEFINITION


class TestFX45Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 45 are properly executed."""

    VERSION = FX45_DEFINITION

    def test_nsIURIChecker_removed(self):
        variations = (
            'request.QueryInterface(Components.interfaces.nsIURIChecker);',
            'IURIChecker = Components.interfaces.nsIURIChecker;'
        )
        for variation in variations:
            self.run_script_for_compat(variation)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            assert len(self.compat_err.warnings) == 1
            assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
            assert (
                self.compat_err.warnings[0]['message']
                == 'The nsIURIChecker interface has been removed')

    def test_identity_interface_changed(self):
        variations = (
            'var gProxyFavIcon = document.getElementById("page-proxy-favicon");',
            'var iconURL = gProxyFavIcon.getAttribute("src");'
        )
        for variation in variations:
            self.run_script_for_compat(variation)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            assert len(self.compat_err.warnings) == 1
            assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
            assert self.compat_err.warnings[0]['message'].startswith(
                'The site identity interface has changed')

    def test_removeAllPages_deprecated(self):
        variations = (
            'historyManager.removeAllPages(); ',
            'historyService.removeAllPages(); '
        )
        for variation in variations:
            self.run_script_for_compat(variation)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            assert len(self.compat_err.warnings) == 1
            assert self.compat_err.warnings[0]['compatibility_type'] == 'warning'
            assert (
                self.compat_err.warnings[0]['message'] ==
                'The removeAllPages function is now deprecated.')



from helper import CompatTestCase
from validator.compat import FX46_DEFINITION


class TestFX46Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 46 are properly executed."""

    VERSION = FX46_DEFINITION

    def test_mTab_renames(self):
        variations = (
            'filter.removeProgressListener(this.mTabListeners[index]);',
            'let tabListener = browser.mTabListeners[selectedIndex];',
            'const filter = this.mTabFilters[index];',
            'var filter = gBrowser.mTabFilters[targetPos];',
        )
        for variation in variations:
            self.run_script_for_compat(variation)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            assert len(self.compat_err.warnings) == 1
            assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
            assert self.compat_err.warnings[0]['message'].startswith(
                'mTabListeners and mTabFilters are now Map objects')



import os
from StringIO import StringIO

import fastchardet
import fnmatch

from validator import decorator
from validator.chromemanifest import ChromeManifest
from validator.constants import PACKAGE_EXTENSION, PACKAGE_LANGPACK
from validator.xpi import XPIManager

from .l10n import dtd, properties


# The threshold that determines the number of entities that must not be
# missing from the package.
L10N_THRESHOLD = 0.35
L10N_SIMILAR_THRESHOLD = 0.9  # For en_US/en_GB kind of stuff

# Only warn about unchanged entities longer than this number of characters.
L10N_LENGTH_THRESHOLD = 3

# To avoid noise, this value ensures that the percent of unchanged entities
# is not inflated due to small numbers of entities.
L10N_MIN_ENTITIES = 18

LOCALE_CACHE = {}


def _list_locales(err, xpi_package=None):
    'Returns a raw list of locales from chrome.manifest'

    chrome = None
    if xpi_package is not None:
        # Handle a reference XPI
        chrome = ChromeManifest(xpi_package.read('chrome.manifest'),
                                'chrome.manifest')
    else:
        # Handle the current XPI
        chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return None

    pack_locales = chrome.get_triples('locale')
    return list(pack_locales)

def _get_locales(err, xpi_package=None, locales=None):
    'Returns a list of locales from the chrome.manifest file.'

    if locales is None:
        locales = _list_locales(err, xpi_package)
    if not locales:
        return None

    processed_locales = {}
    for locale in locales:
        locale_jar = locale['object'].split()

        location = locale_jar[-1]

        # Locales can be bundled in JARs
        jarred = location.startswith('jar:')
        if jarred:
            # We just care about the JAR path
            location = location[4:]
            split_location = location.split('!', 2)
            # Ignore malformed JAR URIs.
            if len(split_location) < 2:
                continue

            path, location = split_location

        locale_desc = {'predicate': locale['predicate'],
                       'target': location,
                       'name': locale_jar[0],
                       'jarred': jarred}
        if jarred:
            locale_desc['path'] = path

        locale_name = '%s:%s' % (locale['predicate'], locale_jar[0])
        if locale_name not in processed_locales:
            processed_locales[locale_name] = locale_desc

    return processed_locales


def _get_locale_manager(err, xpi_package, description,
                        no_cache=False):
    'Returns the XPIManager object for a locale'

    if not description['jarred']:
        return xpi_package

    path = description['path']

    if path in LOCALE_CACHE and not no_cache:
        return LOCALE_CACHE[path]

    if path not in xpi_package:
        # TODO: Pass the filename of the triple's manifest.
        err.warning(('testcases_l10ncompleteness',
                     '_get_locale_manager',
                     'manager_absent'),
                    'Listed locale does not exist',
                    ['A locale JAR is listed in chrome.manifest, but it could '
                     'not be located. Check the spelling and capitalization '
                     'of the path.',
                     'Missing JAR: %s' % path],
                    filename='chrome.manifest')
        return None
    jar = StringIO(xpi_package.read(path))
    locale = XPIManager(jar, mode='r', name=path)

    if not no_cache:
        LOCALE_CACHE[path] = locale
    return locale


@decorator.register_test(tier=4)
def test_xpi(err, xpi_package):
    """Tests an XPI (or JAR, really) for L10n completeness"""

    # Skip over incompatible (or unnecessary) package types.
    if (err.detected_type != PACKAGE_EXTENSION or
        err.is_nested_package):
        return None

    # Don't even both with the test(s) if there's no chrome.manifest.
    if 'chrome.manifest' not in xpi_package:
        return None

    raw_locales = _list_locales(err)

    # We need at least a reference and a target.
    num_locales = len(raw_locales)
    if num_locales < 2:
        if num_locales == 0:
            # TODO: Pass the filename of the triple's manifest.
            err.notice(('testcases_l10ncompleteness',
                        'test_xpi',
                        'no_locales'),
                       'Add-on appears not to be localized',
                       "The add-on doesn't have any locale entries in its "
                       'chrome.manifest file, which suggests that it may '
                       'not be localized.',
                       filename='chrome.manifest')
        return

    locales = _get_locales(err, None, raw_locales)

    # Use the first locale by default
    ref_lang = locales.values()[0]['name']
    # Try to find en-US, as this is where the majority of users is.
    if ref_lang != 'en-US':
        for name, locale in locales.items():
            if locale['name'] == 'en-US':
                ref_lang = 'en-US'
                break

    references = {}
    for locale_name, locale in locales.items():
        if locale['name'] == ref_lang:
            references[locale_name] = locale

    references_locales = {}
    for ref_name, reference in references.items():
        references_locales[ref_name] = \
                _get_locale_manager(err, xpi_package, reference)

    def get_reference(locale):
        for r_name, r in references.items():
            if r['predicate'] == locale['predicate']:
                return r_name, r
        return None, None

    # Loop through the locales and test the valid ones.
    for name, locale in locales.items():
        # Ignore the reference locale
        if name in references:
            continue

        ref_name, reference = get_reference(locale)

        # If we can't find a reference for the current namespace, bail.
        if reference is None or ref_name is None:
            continue

        target_locale = _get_locale_manager(err,
                                            xpi_package,
                                            locale)
        # If we can't find the target locale, just bail.
        if target_locale is None:
            continue

        # Isolate each of the target locales' results.
        results = _compare_packages(references_locales[ref_name],
                                    target_locale,
                                    reference['target'],
                                    locale['target'])
        similar = reference['name'].startswith(locale['name'].split('-')[0])
        _aggregate_results(err, results, locale, similar)


@decorator.register_test(tier=4, expected_type=PACKAGE_LANGPACK)
def test_lp_xpi(err, xpi_package):
    'Tests a language pack for L10n completeness'

    # Don't even both with the test(s) if there's no chrome.manifest.
    if 'chrome.manifest' not in xpi_package:
        return None

    locales = _get_locales(err)

    # Get the reference packages.
    references = []
    support_references = err.get_resource('supports')
    if not support_references:
        references.append('firefox')
        err.info(('testcases_l10ncompleteness',
                  'test_lp_xpi',
                  'missing_app_support'),
                 'Supported app missing in localization completeness.',
                 'While testing in localization comleteness, a list of '
                 'supported applications for the language pack was not found. '
                 'This is likely because there are no listed '
                 '<em:targetApplication> elements in the install.rdf file.')
    else:
        for support in support_references:
            ref_xpi = XPIManager(os.path.join(os.path.dirname(__file__),
                                              'langpacks/%s.xpi' % support))
            ref_xpi.app_name = support
            reference_locales = _get_locales(None, ref_xpi)

            references.append((ref_xpi, reference_locales))

    # Iterate each supported reference package
    for (ref_xpi, ref_locales) in references:
        # Iterate each locale in each supported reference package
        ref_pack = _get_locale_manager(err,
                                       ref_xpi,
                                       {'path': 'en-US.jar',
                                        'jarred': True},
                                       no_cache=True)
        for ref_locale_name in ref_locales:
            ref_locale = ref_locales[ref_locale_name]
            ref_predicate = ref_locale['predicate']
            corresp_locales = [locales[name] for name
                               in locales
                               if locales[name]['predicate'] == ref_predicate]
            # If we found no matching locale, then it's missing from the pack
            if not corresp_locales:
                err.warning(('testcases_l10ncompleteness',
                             'test_lp_xpi',
                             'find_corresponding_locale'),
                            'Could not find corresponding locale',
                            ['A locale was found in the reference package, '
                             'however it was not found in the target package.',
                             'Missing locale: %s' % ref_predicate],
                            filename='chrome.manifest')
                continue

            target_locale = corresp_locales[0]
            target_pack = _get_locale_manager(err,
                                              xpi_package,
                                              target_locale)
            if target_pack is None:
                continue

            results = _compare_packages(reference=ref_pack,
                                        target=target_pack,
                                        ref_base=ref_locale['target'],
                                        locale_base=target_locale['target'])

            # Report the findings after each supported app's locale
            _aggregate_results(err, results, target_locale)

    # Clear the cache at the end of the test
    LOCALE_CACHE.clear()


def _compare_packages(reference, target, ref_base='', locale_base=''):
    'Compares two L10n-compatible packages to one another.'

    tar_files = target.package_contents()

    results = []
    total_entities = 0

    ref_base = ref_base.lstrip('/')
    locale_base = locale_base.lstrip('/')

    l10n_docs = ('dtd', 'properties', 'xhtml', 'ini', 'inc')
    parsable_docs = ('dtd', 'properties')

    for name in reference:

        entity_count = 0

        # Skip directory entries.
        if name.endswith('/'):  # pragma: no cover
            continue

        # Ignore files not considered reference files.
        if ref_base and not name.startswith(ref_base):
            continue

        extension = name.split('.')[-1]
        if extension not in l10n_docs:
            continue
        parsable = extension in parsable_docs

        if parsable:
            ref_doc = _parse_l10n_doc(name,
                                      reference.read(name),
                                      no_encoding=True)
        else:
            ref_doc = ()

        tar_name = locale_base + name[len(ref_base):]
        if tar_name not in tar_files:
            results.append({'type': 'missing_files',
                            'entities': len(ref_doc),
                            'filename': tar_name})
            continue

        if not parsable:
            continue

        tar_doc = _parse_l10n_doc(tar_name, target.read(tar_name))

        if not tar_doc.expected_encoding:
            results.append({'type': 'unexpected_encoding',
                            'filename': tar_name,
                            'expected_encoding': tar_doc.suitable_encoding})

        missing_entities = []
        unchanged_entities = []

        for rname, rvalue, rline in ref_doc.items:
            entity_count += 1

            if rname not in tar_doc.entities:
                missing_entities.append(rname)
                continue

            if (not rname.startswith('pref.timezone.') and
                rvalue == tar_doc.entities[rname] and
                len(rvalue) > L10N_LENGTH_THRESHOLD and
                not fnmatch.fnmatch(rvalue, 'http*://*')):

                unchanged_entities.append((rname, rline))
                continue

        if missing_entities:
            results.append({'type': 'missing_entities',
                            'entities': len(missing_entities),
                            'filename': tar_name,
                            'missing_entities': missing_entities})
        if unchanged_entities:
            results.append({'type': 'unchanged_entity',
                            'entities': len(unchanged_entities),
                            'filename': tar_name,
                            'unchanged_entities': unchanged_entities})

        results.append({'type': 'file_entity_count',
                        'filename': tar_name,
                        'entities': entity_count})

        total_entities += entity_count

    results.append({'type': 'total_entities',
                    'entities': total_entities})
    return results


def _parse_l10n_doc(name, doc, no_encoding=False):
    'Parses an L10n document.'

    extension = name.split('.')[-1].lower()

    handlers = {'dtd': dtd.DTDParser,
                'properties': properties.PropertiesParser}
    # These are expected encodings for the various files.
    handler_formats = ('ASCII', 'UTF_8')
    if extension not in handlers:
        return None

    wrapper = StringIO(doc)
    loc_doc = handlers[extension](wrapper)

    # Allow the parse to specify files to skip for encoding checks
    if not no_encoding:
        try:
            # This is much faster than fastchardet, and succeeds more often
            # than fails.
            doc.decode('utf-8')
            encoding = 'UTF_8'
        except UnicodeDecodeError:
            encoding = fastchardet.detect(doc)['encoding'].upper()
        loc_doc.expected_encoding = encoding in handler_formats
        loc_doc.suitable_encoding = handler_formats

    return loc_doc


def _aggregate_results(err, results, locale, similar=False, base='en-US'):
    """Compiles the errors and warnings in the L10n results list into
    error bundler errors and warnings."""

    total_entities = 0
    unchanged_entity_list = {}
    entity_count = {}
    unexpected_encodings = []

    for ritem in results:
        if 'filename' in ritem:
            rfilename = ritem['filename']

        rtype = ritem['type']
        if rtype == 'missing_files':
            err.warning(('testcases_l10ncompleteness',
                        '_aggregate_results',
                        'missing_file'),
                       'Missing translation file',
                       ['Localizations must include a translated copy of each '
                        'file in the reference locale. The required files may '
                        'vary from target application to target application.',
                        'Missing translation file: %s' % rfilename],
                      locale['target'])
        elif rtype == 'missing_entities':
            err.warning(('testcases_l10ncompleteness',
                         '_aggregate_results',
                         'missing_translation_entity'),
                        'Missing translation entity',
                        ['Localizations must include a translated copy of each '
                         'entity from each file in the reference locale. The '
                         'required files may vary from target application to '
                         'target application.',
                         'Missing Entities: %s' %
                             ', '.join(ritem['missing_entities'])],
                        [locale['target'], rfilename])
        elif rtype == 'unchanged_entity':
            filename = ritem['filename']
            if filename not in unchanged_entity_list:
                unchanged_entity_list[filename] = {'count': 0,
                                                   'entities': []}
            unchanged = unchanged_entity_list[filename]
            unchanged['count'] += ritem['entities']
            unchanged['entities'].extend(ritem['unchanged_entities'])
        elif rtype == 'total_entities':
            total_entities += ritem['entities']
        elif rtype == 'file_entity_count':
            entity_count[ritem['filename']] = ritem['entities']
        elif rtype == 'unexpected_encoding':
            unexpected_encodings.append(
                    (ritem['filename'],
                     ', '.join(ritem['expected_encoding'])))

    # Determine the locale's filename argument in advance since we'll use it
    # for every message
    locale_filename = [locale['target']]
    if locale['jarred']:
        locale_filename.append(locale['path'].lstrip('/'))

    agg_unchanged = []
    if not similar:
        unchanged_percentage = L10N_THRESHOLD
    else:
        unchanged_percentage = L10N_SIMILAR_THRESHOLD
    for name, count in entity_count.items():
        if name not in unchanged_entity_list or \
           count == 0:
            continue

        unchanged = unchanged_entity_list[name]
        total_adjusted = max(count, L10N_MIN_ENTITIES)
        percentage = float(unchanged['count']) / float(total_adjusted)
        if percentage >= unchanged_percentage:
            agg_unchanged.append(
                    '%s: %d/%d entities unchanged (%s) at %d percent' %
                    (name,
                     unchanged['count'],
                     count,
                     ', '.join(['%s (%d)' % (e, line)
                                for e, line
                                in unchanged['entities']]),
                     percentage * 100))

    if agg_unchanged:
        err.notice(('testcases_l10ncompleteness',
                    '_aggregate_results',
                    'unchanged_entities'),
                   'Unchanged translation entities',
                   ['Localizations must include a translated copy of each '
                    'entity from each file in the reference locale. These '
                    'translations SHOULD differ from the localized text in '
                    'the reference package.',
                    agg_unchanged],
                   locale_filename)

    if unexpected_encodings:
        # Compile all of the encoding errors into one nice warning.
        compilation = ['Detected files:']
        for target in unexpected_encodings:
            compilation.append('%s\n (expected: %s)' % target)

        err.warning(('testcases_l10ncompleteness',
                     '_aggregate_results',
                     'unexpected_encodings'),
                    'Unexpected encodings in locale files',
                    ['Localization files were encountered that used encodings '
                     'that are not characteristic of those types of files.',
                     '\n'.join(compilation),
                     'Localization files with the wrong encoding can cause '
                     'issues with locales that include non-ASCII characters.'],
                    locale_filename)




import codecs
import re

# Many thanks to nmaier for inspiration and code in this module

UNICODE_BOMS = [
    (codecs.BOM_UTF8, 'utf-8'),
    (codecs.BOM_UTF32_LE, 'utf-32-le'),
    (codecs.BOM_UTF32_BE, 'utf-32-be'),
    (codecs.BOM_UTF16_LE, 'utf-16-le'),
    (codecs.BOM_UTF16_BE, 'utf-16-be'),
]

COMMON_ENCODINGS = ('latin_1', 'utf-16')

# Matches any non-ASCII characters, and any unprintable characters in the
# 7-bit ASCII range. Accepts tab, return, newline, and any other character
# code above 0x20 which fits in 7 bits.
NON_ASCII_FILTER = re.compile(r'[^\t\r\n\x20-\x7f]+')


def decode(data):
    """
    Decode data employing some charset detection and including unicode BOM
    stripping.
    """

    if isinstance(data, unicode):
        return data

    # Detect standard unicode BOMs.
    for bom, encoding in UNICODE_BOMS:
        if data.startswith(bom):
            return data[len(bom):].decode(encoding, errors='ignore')

    # Try straight UTF-8.
    try:
        return data.decode('utf-8')
    except UnicodeDecodeError:
        pass

    # Test for various common encodings.
    for encoding in COMMON_ENCODINGS:
        try:
            return data.decode(encoding)
        except UnicodeDecodeError:
            pass

    # Anything else gets filtered.
    return NON_ASCII_FILTER.sub('', data).decode('ascii', errors='replace')



__version__ = '1.10.70'


class ValidationTimeout(Exception):
    """Validation has timed out.

    May be replaced by the exception type raised by an external timeout
    handler when run in a server environment."""

    def __init__(self, timeout):
        self.timeout = timeout

    def __str__(self):
        return 'Validation timeout after %d seconds' % self.timeout



from validator.constants import PACKAGE_THEME
from validator.contextgenerator import ContextGenerator
from validator.testcases.javascript import traverser
from validator.testcases.javascript.jsshell import get_tree


def test_js_file(err, filename, data, line=0, context=None, pollutable=False):
    'Test a JS file by parsing and analyzing its tokens.'

    if err.detected_type == PACKAGE_THEME:
        err.warning(
                err_id=('testcases_scripting',
                        'test_js_file',
                        'theme_js'),
                warning='JS run from full theme',
                description='Themes should not contain executable code.',
                filename=filename,
                line=line)

    before_tier = None
    # Set the tier to 4 (Security Tests)
    if err is not None:
        before_tier = err.tier
        err.set_tier(3)

    tree = get_tree(data, filename=filename, err=err)
    if not tree:
        if before_tier:
            err.set_tier(before_tier)
        return

    # Generate a context if one is not available.
    if context is None:
        context = ContextGenerator(data)

    t = traverser.Traverser(err, filename, line, context=context,
                            is_jsm=(filename.endswith('.jsm') or
                                    'EXPORTED_SYMBOLS' in data))
    t.pollutable = pollutable
    t.run(tree)

    # Reset the tier so we don't break the world
    if err is not None:
        err.set_tier(before_tier)


def test_js_snippet(err, data, filename, line=0, context=None):
    'Process a JS snippet by passing it through to the file tester.'

    if not data:
        return

    # Wrap snippets in a function to prevent the parser from freaking out
    # when return statements exist without a corresponding function.
    data = '(function(){%s\n})()' % data

    test_js_file(err, filename, data, line, context, pollutable=False)



import os.path

from validator.contextgenerator import ContextGenerator


class ChromeManifest(object):
    """This class enables convenient parsing and iteration of
    chrome.manifest files."""

    def __init__(self, data, path):
        self.context = ContextGenerator(data)
        self.lines = data.split('\n')

        # Extract the data from the triples in the manifest
        triples = []
        counter = 0

        for line in self.lines:
            line = line.strip()

            counter += 1

            # Skip weird lines.
            if line.startswith('#'):
                continue

            triple = line.split(None, 2)
            if not triple:
                continue
            elif len(triple) == 2:
                triple.append('')
            if len(triple) < 3:
                continue

            triples.append({'subject': triple[0],
                            'predicate': triple[1],
                            'object': triple[2],
                            'line': counter,
                            'filename': path,
                            'context': self.context})

        self.triples = triples

    def get_value(self, subject=None, predicate=None, object_=None):
        """Returns the first triple value matching the given subject,
        predicate, and/or object"""

        for triple in self.triples:

            # Filter out non-matches
            if (subject and triple['subject'] != subject) or \
               (predicate and triple['predicate'] != predicate) or \
               (object_ and triple['object'] != object_):  # pragma: no cover
                continue

            # Return the first found.
            return triple

        return None

    def get_objects(self, subject=None, predicate=None):
        """Returns a generator of objects that correspond to the
        specified subjects and predicates."""

        for triple in self.triples:

            # Filter out non-matches
            if ((subject and triple['subject'] != subject) or
                (predicate and triple['predicate'] != predicate)):
                continue

            yield triple['object']

    def get_triples(self, subject=None, predicate=None, object_=None):
        """Returns triples that correspond to the specified subject,
        predicates, and objects."""

        for triple in self.triples:

            # Filter out non-matches
            if subject is not None and triple['subject'] != subject:
                continue
            if predicate is not None and triple['predicate'] != predicate:
                continue
            if object_ is not None and triple['object'] != object_:
                continue

            yield triple

    def get_applicable_overlays(self, error_bundle):
        """
        Given an error bundle, a list of overlays that are present in the
        current package or subpackage are returned.
        """

        content_paths = self.get_triples(subject='content')
        if not content_paths:
            return set()

        # Create some variables that will store where the applicable content
        # instruction path references and where it links to.
        chrome_path = ''
        content_root_path = '/'

        # Look through each of the listed packages and paths.
        for path in content_paths:
            chrome_name = path['predicate']
            if not path['object']:
                continue
            path_location = path['object'].strip().split()[0]

            # Handle jarred paths differently.
            if path_location.startswith('jar:'):
                if not error_bundle.is_nested_package:
                    continue

                # Parse out the JAR and it's location within the chrome.
                split_jar_url = path_location[4:].split('!', 2)
                # Ignore invalid/unsupported JAR URLs.
                if len(split_jar_url) != 2:
                    continue

                # Unpack the JAR URL.
                jar_path, package_path = split_jar_url

                # Ignore the instruction if the JAR it points to doesn't match
                # up with the current subpackage tree.
                if jar_path != error_bundle.package_stack[0]:
                    continue
                chrome_path = self._url_chunk_join(chrome_name, package_path)
                # content_root_path stays at the default: /

                break
            else:
                # If we're in a subpackage, a content instruction referring to
                # the root of the package obviously doesn't apply.
                if error_bundle.is_nested_package:
                    continue

                chrome_path = self._url_chunk_join(chrome_name, 'content')
                content_root_path = '/%s/' % path_location.strip('/')
                break

        if not chrome_path:
            return set()

        applicable_overlays = set()
        chrome_path = 'chrome://%s' % self._url_chunk_join(chrome_path + '/')

        for overlay in self.get_triples(subject='overlay'):
            if not overlay['object']:
                error_bundle.error(
                    err_id=('chromemanifest', 'get_applicable_overalys',
                            'object'),
                    error='Overlay instruction missing a property.',
                    description='When overlays are registered in a chrome '
                                'manifest file, they require a namespace and '
                                'a chrome URL at minimum.',
                    filename=overlay['filename'],
                    line=overlay['line'],
                    context=self.context)  #TODO(basta): Update this!
                continue
            overlay_url = overlay['object'].split()[0]
            if overlay_url.startswith(chrome_path):
                overlay_relative_path = overlay_url[len(chrome_path):]
                applicable_overlays.add('/%s' %
                        self._url_chunk_join(content_root_path,
                                             overlay_relative_path))

        return applicable_overlays

    def reverse_lookup(self, state, path):
        """
        Returns a chrome URL for a given path, given the current package depth
        in an error bundle.

        State may either be an error bundle or the actual package stack.
        """

        # Make sure the path starts with a forward slash.
        if not path.startswith('/'):
            path = '/%s' % path

        # If the state is an error bundle, extract the package stack.
        if not isinstance(state, list):
            state = state.package_stack

        content_paths = self.get_triples(subject='content')
        for content_path in content_paths:
            chrome_name = content_path['predicate']
            if not content_path['object']:
                continue
            path_location = content_path['object'].split()[0]

            if path_location.startswith('jar:'):
                if not state:
                    continue

                # Parse out the JAR and it's location within the chrome.
                split_jar_url = path_location[4:].split('!', 2)
                # Ignore invalid/unsupported JAR URLs.
                if len(split_jar_url) != 2:
                    continue

                # Unpack the JAR URL.
                jar_path, package_path = split_jar_url

                if jar_path != state[0]:
                    continue

                return 'chrome://%s' % self._url_chunk_join(chrome_name,
                                                            package_path,
                                                            path)
            else:
                if state:
                    continue

                path_location = '/%s/' % path_location.strip('/')
                rel_path = os.path.relpath(path, path_location)

                if rel_path.startswith('../') or rel_path == '..':
                    continue

                return 'chrome://%s' % self._url_chunk_join(chrome_name,
                                                            rel_path)

        return None

    def _url_chunk_join(self, *args):
        """Join the arguments together to form a predictable URL chunk."""
        # Strip slashes from either side of each path piece.
        pathlets = map(lambda s: s.strip('/'), args)
        # Remove empty pieces.
        pathlets = filter(None, pathlets)
        url = '/'.join(pathlets)
        # If this is a directory, add a trailing slash.
        if args[-1].endswith('/'):
            url = '%s/' % url
        return url




from validator.constants import *


def detect_type(err, install_rdf=None, xpi_package=None):
    """Determines the type of add-on being validated based on
    install.rdf, file extension, and other properties."""

    # The types in the install.rdf don't pair up 1:1 with the type
    # system that we're using for expectations and the like. This is
    # to help translate between the two.
    translated_types = {'2': PACKAGE_EXTENSION,
                        '4': PACKAGE_THEME,
                        '8': PACKAGE_LANGPACK,
                        '32': PACKAGE_MULTI,
                        '64': PACKAGE_DICTIONARY,
                        # New "experiment" types: see bug 1220097 and
                        # https://github.com/mozilla/addons-server/issues/3315
                        '128': PACKAGE_EXTENSION,
                        '256': PACKAGE_EXTENSION,}

    # If we're missing our install.rdf file, we can try to make some
    # assumptions.
    if install_rdf is None:
        types = {'xpi': PACKAGE_DICTIONARY}

        err.notice(('typedetection',
                    'detect_type',
                    'missing_install_rdf'),
                   'install.rdf was not found.',
                   'The type should be determined by install.rdf if present. '
                   "If it isn't, we still need to know the type.")

        # If we know what the file type might be, return it.
        if xpi_package.extension in types:
            return types[xpi_package.extension]
        # Otherwise, we're out of luck :(
        else:
            return None

    # Attempt to locate the <em:type> node in the RDF doc.
    type_uri = install_rdf.uri('type')
    type_ = install_rdf.get_object(None, type_uri)

    # Dictionaries are weird too, they might not have the obligatory
    # em:type. We can assume that if they have a /dictionaries/ folder,
    # they are a dictionary because even if they aren't, dictionaries
    # have an extraordinarily strict set of rules and file filters that
    # must be passed. It's so crazy secure that it's cool if we use it
    # as kind of a fallback.

    if any(file_ for file_ in xpi_package if
               file_.startswith('dictionaries/')):
        if type_ != '64':
            err.error(('typedetection',
                       'dictionary_valid_type',
                       'invalid_em_type'),
                      'Invalid <em:type> value.',
                      'The package appears to be a dictionary but does not have '
                      'the correct <em:type> set in the install manifest.')
        return PACKAGE_DICTIONARY

    if type_ is not None:
        if type_ in translated_types:
            err.save_resource('is_multipackage', type_ == '32', pushable=True)
            # Make sure we translate back to the normalized version
            return translated_types[type_]
        else:
            err.error(('typedetection',
                       'detect_type',
                       'invalid_em_type'),
                      'Invalid <em:type> value.',
                      'The only valid values for <em:type> are 2, 4, 8, and '
                      '32. Any other values are either invalid or deprecated.',
                      'install.rdf')
            return
    else:
        err.notice(
            err_id=('typedetection',
                    'detect_type',
                    'no_em:type'),
            notice='No <em:type> element found in install.rdf',
            description="It isn't always required, but it is the most reliable "
                        'method for determining add-on type.',
            filename='install.rdf')

    # There's no type element, so the spec says that it's either a
    # theme or an extension. At this point, we know that it isn't
    # a dictionary, language pack, or multiple extension pack.
    extensions = {'jar': '4',
                  'xpi': '2'}

    # If the package's extension is listed in the [tiny] extension
    # dictionary, then just return that. We'll validate against that
    # add-on type's layout later. Better to false positive than to false
    # negative.
    if xpi_package.extension in extensions:
        # Make sure it gets translated back to the normalized version
        install_rdf_type = extensions[xpi_package.extension]
        return translated_types[install_rdf_type]

